
#include "mainwindow.h"
#include "math.h"
#include "ui_mainwindow.h"
#include <QtGui>

/*******************************************************************************
Open the training dataset.
        posdirectory: Directory containing face images
        negdirectory: Directory containing non-face images
        trainingData: Array used to store the data
        trainingLabel: Label assigned to training data (1 = face, 0 = non-face)
        numTrainingExamples: Number of training examples
        patchSize: Size of training patches
*******************************************************************************/
void MainWindow::OpenDataSet(QDir posdirectory, QDir negdirectory, double *trainingData, int *trainingLabel, int numTrainingExamples, int patchSize)
{
    int i, c, r;
    QStringList imgNames;
    QImage inImage;
    QRgb pixel;

    imgNames = posdirectory.entryList();

    int idx = 0;

    for(i=0;i<imgNames.length();i++)
        if(idx < numTrainingExamples/2)
    {
        inImage.load(posdirectory.absolutePath() + "\\" + imgNames.at(i));

        if(!(inImage.isNull()))
        {
            for(r=0;r<patchSize;r++)
                for(c=0;c<patchSize;c++)
                {
                    pixel = inImage.pixel(c, r);
                    trainingData[idx*patchSize*patchSize + r*patchSize + c] = (double) qGreen(pixel);
                }

            trainingLabel[idx] = 1;

            idx++;
        }
    }

    imgNames = negdirectory.entryList();

    for(i=0;i<imgNames.length();i++)
        if(idx < numTrainingExamples)
    {
        inImage.load(negdirectory.absolutePath() + "\\" + imgNames.at(i));

        if(!(inImage.isNull()))
        {
            for(r=0;r<patchSize;r++)
                for(c=0;c<patchSize;c++)
                {
                    pixel = inImage.pixel(c, r);
                    trainingData[idx*patchSize*patchSize + r*patchSize + c] = (double) qGreen(pixel);
                }

            trainingLabel[idx] = 0;

            idx++;
        }
    }
}

/*******************************************************************************
Display example patches from training dataset.
        displayImage: Display image
        trainingData: Array used to store the data
        trainingLabel: Label assigned to training data (1 = face, 0 = non-face)
        numTrainingExamples: Number of training examples
        patchSize: Size of training patches
*******************************************************************************/
void MainWindow::DisplayTrainingDataset(QImage *displayImage, double *trainingData, int *trainingLabel, int numTrainingExamples, int patchSize)
{
    int w = displayImage->width();
    int h = displayImage->height();
    int r, c;
    int rOffset = 0;
    int cOffset = 0;
    bool inBounds = true;
    int ct = 0;

    while(inBounds)
    {
        int idx = rand()%numTrainingExamples;

        for(r=0;r<patchSize;r++)
            for(c=0;c<patchSize;c++)
            {
                if(trainingLabel[idx] == 1)
                {
                    int val = (int) trainingData[idx*patchSize*patchSize + r*patchSize + c];
                    displayImage->setPixel(c + cOffset, r + rOffset, qRgb(val, val, val));

                }
                else
                {
                    int val = (int) trainingData[idx*patchSize*patchSize + r*patchSize + c];
                    displayImage->setPixel(c + cOffset, r + rOffset, qRgb(val, val, val));
                }
            }

        cOffset += patchSize;

        if(cOffset + patchSize >= w)
        {
            cOffset = 0;
            rOffset += patchSize;

            if(rOffset + patchSize >= h)
                inBounds = false;
        }

        ct++;
    }
}

/*******************************************************************************
Save the computed AdaBoost classifier.
        fileName: Name of file
*******************************************************************************/
void MainWindow::SaveClassifier(QString fileName)
{
   int i, j;
   FILE *out;

   out = fopen(fileName.toAscii(), "w");

   fprintf(out, "%d\n", m_NumWeakClassifiers);

   for(i=0;i<m_NumWeakClassifiers;i++)
   {
       fprintf(out, "%d\n", m_WeakClassifiers[i].m_NumBoxes);

       for(j=0;j<m_WeakClassifiers[i].m_NumBoxes;j++)
           fprintf(out, "%lf\t%lf\t%lf\t%lf\t%lf\n", m_WeakClassifiers[i].m_BoxSign[j], m_WeakClassifiers[i].m_Box[j][0][0], m_WeakClassifiers[i].m_Box[j][0][1],
                   m_WeakClassifiers[i].m_Box[j][1][0], m_WeakClassifiers[i].m_Box[j][1][1]);

       fprintf(out, "%lf\n", m_WeakClassifiers[i].m_Area);
       fprintf(out, "%lf\n", m_WeakClassifiers[i].m_Polarity);
       fprintf(out, "%lf\n", m_WeakClassifiers[i].m_Threshold);
       fprintf(out, "%lf\n", m_WeakClassifiers[i].m_Weight);
   }

   fclose(out);
}

/*******************************************************************************
Open the computed AdaBoost classifier.
        fileName: Name of file
*******************************************************************************/
void MainWindow::OpenClassifier(QString fileName)
{
    int i, j;
    FILE *in;

    in = fopen(fileName.toAscii(), "r");

    fscanf(in, "%d\n", &m_NumWeakClassifiers);
    m_WeakClassifiers = new CWeakClassifiers [m_NumWeakClassifiers];

    for(i=0;i<m_NumWeakClassifiers;i++)
    {
        fscanf(in, "%d\n", &(m_WeakClassifiers[i].m_NumBoxes));
        m_WeakClassifiers[i].m_Box = new double [m_WeakClassifiers[i].m_NumBoxes][2][2];
        m_WeakClassifiers[i].m_BoxSign = new double [m_WeakClassifiers[i].m_NumBoxes];

        for(j=0;j<m_WeakClassifiers[i].m_NumBoxes;j++)
            fscanf(in, "%lf\t%lf\t%lf\t%lf\t%lf\n", &(m_WeakClassifiers[i].m_BoxSign[j]), &(m_WeakClassifiers[i].m_Box[j][0][0]), &(m_WeakClassifiers[i].m_Box[j][0][1]),
                    &(m_WeakClassifiers[i].m_Box[j][1][0]), &(m_WeakClassifiers[i].m_Box[j][1][1]));

        fscanf(in, "%lf\n", &(m_WeakClassifiers[i].m_Area));
        fscanf(in, "%lf\n", &(m_WeakClassifiers[i].m_Polarity));
        fscanf(in, "%lf\n", &(m_WeakClassifiers[i].m_Threshold));
        fscanf(in, "%lf\n", &(m_WeakClassifiers[i].m_Weight));
    }

    fclose(in);

}

/*******************************************************************************
Display the Haar wavelets for the classifier.
        displayImage: Display image
        weakClassifiers: The weak classifiers used in AdaBoost
        numWeakClassifiers: Number of weak classifiers
*******************************************************************************/
void MainWindow::DisplayClassifiers(QImage *displayImage, CWeakClassifiers *weakClassifiers, int numWeakClassifiers)
{
    int w = displayImage->width();
    int h = displayImage->height();
    int i, j, r, c;
    int rOffset = 0;
    int cOffset = 0;
    int size = 50;
    bool inBounds = true;

    displayImage->fill(qRgb(0,0,0));

    for(i=0;i<numWeakClassifiers & inBounds;i++)
    {
        for(r=0;r<size;r++)
            for(c=0;c<size;c++)
            {
                 displayImage->setPixel(c + cOffset, r + rOffset, qRgb(128, 128, 128));
            }

        for(j=0;j<weakClassifiers[i].m_NumBoxes;j++)
            for(r=(int) ((double) size*weakClassifiers[i].m_Box[j][0][1]);r<(int) ((double) size*weakClassifiers[i].m_Box[j][1][1]);r++)
                for(c=(int) ((double) size*weakClassifiers[i].m_Box[j][0][0]);c<(int) ((double) size*weakClassifiers[i].m_Box[j][1][0]);c++)
                {
                    if(weakClassifiers[i].m_BoxSign[j] > 0.0)
                        displayImage->setPixel(c + cOffset, r + rOffset, qRgb(255, 255, 255));
                    else
                        displayImage->setPixel(c + cOffset, r + rOffset, qRgb(0, 0, 0));
                }

        cOffset += size+1;

        if(cOffset + size >= w)
        {
            cOffset = 0;
            rOffset += size + 1;

            if(rOffset + size >= h)
                inBounds = false;
        }
    }
}

/*******************************************************************************
Display the integral image
        displayImage: Display image
        integralImage: Output integral image
        w, h: Width and height of image
*******************************************************************************/
void MainWindow::DisplayIntegralImage(QImage *displayImage, double *integralImage, int w, int h)
{
    int r, c;
    double maxVal = integralImage[(h-1)*w + w-1];

    for(r=0;r<h;r++)
        for(c=0;c<w;c++)
        {
            int val = (int) (255.0*integralImage[r*w + c]/maxVal);

            displayImage->setPixel(c, r, qRgb(val, val, val));
        }
}

/*******************************************************************************
Randomly initialize the candidate weak classifiers.
        weakClassifiers: Candidate weak classifiers
        numWeakClassifiers: Number of candidate weak classifiers
*******************************************************************************/
void MainWindow::InitializeFeatures(CWeakClassifiers *weakClassifiers, int numWeakClassifiers)
{
    int i;

    for(i=0;i<numWeakClassifiers;i++)
    {
        double x, y, w, h;

        // We don't know these values yet, so just initialize to 0
        weakClassifiers[i].m_Polarity = 0.0;
        weakClassifiers[i].m_Threshold = 0.0;
        weakClassifiers[i].m_Weight = 0.0;

        // The Haar wavelet's corners can range in the area of 0.02 to 0.98, with a minimum size of 0.25
        // We limit the range to [0.02, 0.98], instead of [0, 1] so we don't need to worry about checking
        // out of bounds errors later on, i.e. in the BilinearInterpolation function.

        // x position of box and width
        w = 0.25 + 0.71*(double) rand()/(double) RAND_MAX;
        x = 0.02 + (0.96 - w)*(double) rand()/(double) RAND_MAX;

        // y position of box and height
        h = 0.25 + 0.71*(double) rand()/(double) RAND_MAX;
        y = 0.02 + (0.96 - h)*(double) rand()/(double) RAND_MAX;

        int boxType = rand()%3;

        if(boxType == 0)
        {
            // Vertical boxes
            weakClassifiers[i].m_NumBoxes = 2;
            weakClassifiers[i].m_Box = new double [weakClassifiers[i].m_NumBoxes][2][2];
            weakClassifiers[i].m_BoxSign = new double [weakClassifiers[i].m_NumBoxes];

            weakClassifiers[i].m_BoxSign[0] = 1.0;
            weakClassifiers[i].m_Box[0][0][0] = x;
            weakClassifiers[i].m_Box[0][0][1] = y;
            weakClassifiers[i].m_Box[0][1][0] = x + w/2;
            weakClassifiers[i].m_Box[0][1][1] = y + h;

            weakClassifiers[i].m_BoxSign[1] = -1.0;
            weakClassifiers[i].m_Box[1][0][0] = x + w/2;
            weakClassifiers[i].m_Box[1][0][1] = y;
            weakClassifiers[i].m_Box[1][1][0] = x + w;
            weakClassifiers[i].m_Box[1][1][1] = y + h;
        }

        if(boxType == 1)
        {
            // 2 Horizontal boxes
            weakClassifiers[i].m_NumBoxes = 2;
            weakClassifiers[i].m_Box = new double [weakClassifiers[i].m_NumBoxes][2][2];
            weakClassifiers[i].m_BoxSign = new double [weakClassifiers[i].m_NumBoxes];

            weakClassifiers[i].m_BoxSign[0] = 1.0;
            weakClassifiers[i].m_Box[0][0][0] = x;
            weakClassifiers[i].m_Box[0][0][1] = y;
            weakClassifiers[i].m_Box[0][1][0] = x + w;
            weakClassifiers[i].m_Box[0][1][1] = y + h/2;

            weakClassifiers[i].m_BoxSign[1] = -1.0;
            weakClassifiers[i].m_Box[1][0][0] = x;
            weakClassifiers[i].m_Box[1][0][1] = y + h/2;
            weakClassifiers[i].m_Box[1][1][0] = x + w;
            weakClassifiers[i].m_Box[1][1][1] = y + h;
        }

        if(boxType == 2)
        {
            // 3 Vertical boxes
            weakClassifiers[i].m_NumBoxes = 3;
            weakClassifiers[i].m_Box = new double [weakClassifiers[i].m_NumBoxes][2][2];
            weakClassifiers[i].m_BoxSign = new double [weakClassifiers[i].m_NumBoxes];

            weakClassifiers[i].m_BoxSign[0] = 1.0;
            weakClassifiers[i].m_Box[0][0][0] = x;
            weakClassifiers[i].m_Box[0][0][1] = y;
            weakClassifiers[i].m_Box[0][1][0] = x + w/3;
            weakClassifiers[i].m_Box[0][1][1] = y + h;

            weakClassifiers[i].m_BoxSign[1] = -2.0;
            weakClassifiers[i].m_Box[1][0][0] = x + w/3;
            weakClassifiers[i].m_Box[1][0][1] = y;
            weakClassifiers[i].m_Box[1][1][0] = x + 2*w/3;
            weakClassifiers[i].m_Box[1][1][1] = y + h;

            weakClassifiers[i].m_BoxSign[2] = 1.0;
            weakClassifiers[i].m_Box[2][0][0] = x + 2*w/3;
            weakClassifiers[i].m_Box[2][0][1] = y;
            weakClassifiers[i].m_Box[2][1][0] = x + w;
            weakClassifiers[i].m_Box[2][1][1] = y + h;
        }

        weakClassifiers[i].m_Area = w*h;
    }
}

/*******************************************************************************
Simple helper function to convert from RGB to double.
        image: Input image
        dImage: Output double image
        w, h: Image width and height
*******************************************************************************/
void MainWindow::ConvertColorToDouble(QImage image, double *dImage, int w, int h)
{
    QRgb pixel;
    int r, c;

    for(r=0;r<h;r++)
        for(c=0;c<w;c++)
        {
            pixel = image.pixel(c, r);
            dImage[r*w + c] = qGreen(pixel);
        }
}

/*******************************************************************************
Compute all of the features for the training dataset.
        trainingData: Array used to store the data
        features: Array holding feature values
        numTrainingExamples: Number of training examples
        patchSize: Size of training patches
        weakClassifiers: Candidate weak classifiers
        numWeakClassifiers: Number of candidate weak classifiers
*******************************************************************************/
void MainWindow::ComputeTrainingSetFeatures(double *trainingData, double *features,
                                int numTrainingExamples, int patchSize, CWeakClassifiers *weakClassifiers, int numWeakClassifiers)
{
    int i;
    double *integralImage = new double [patchSize*patchSize];

    for(i=0;i<numTrainingExamples;i++)
    {
        // Compute features for training examples

        // First compute the integral image for each patch
        IntegralImage(&(trainingData[i*patchSize*patchSize]), integralImage, patchSize, patchSize);

        // Compute the Haar wavelets
        ComputeFeatures(integralImage, 0, 0, patchSize, &(features[i*numWeakClassifiers]), weakClassifiers, numWeakClassifiers, patchSize);
    }


    // We shouldn't need the training data anymore so let's delete it.
    delete [] trainingData;

    delete [] integralImage;
}

/*******************************************************************************
Display the computed features (green = faces, red = background)
        displayImage: Display image
        features: Array holding feature values
        trainingLabel: Label assigned to training data (1 = face, 0 = non-face)
        numFeatures: Number of features
        numTrainingExamples: Number of training examples
*******************************************************************************/
void MainWindow::DisplayFeatures(QImage *displayImage, double *features, int *trainingLabel, int numFeatures, int numTrainingExamples)
{
    int r, c;
    int w = displayImage->width();
    int h = displayImage->height();
    int posCt = 0;
    int negCt = 0;

    double mean = 0.0;
    double meanCt = 0.0;

	// Clear image
	for(r = 0; r < displayImage->height(); r++)
	{
		for(c = 0; c < displayImage->width(); c++)
		{
			displayImage->setPixel(c, r, qRgb(0, 0, 0));
		}
	}


    for(r=0;r<numTrainingExamples;r+=10)
    {
        for(c=0;c<numFeatures;c++)
        {
            mean += fabs(features[r*numFeatures + c]);
            meanCt++;
        }
    }

    mean /= meanCt;

    for(r=0;r<numTrainingExamples;r++)
    {
        if(trainingLabel[r] == 1 && posCt < h/2)
        {
            for(c=0;c<numFeatures;c++)
                if(c < w)
            {
                int val = 255.0*(features[r*numFeatures + c]/(4.0*mean)) + 128.0;
                val = min(255, max(0, val));

                displayImage->setPixel(c, posCt, qRgb(0, val, 0));
            }

            posCt++;
        }

        if(trainingLabel[r] == 0 && negCt < h/2)
        {
            for(c=0;c<numFeatures;c++)
                if(c < w)
            {
                int val = 255.0*(features[r*numFeatures + c]/(4.0*mean)) + 128.0;
                val = min(255, max(0, val));

                displayImage->setPixel(c, negCt + h/2, qRgb(val, 0, 0));
            }

            negCt++;
        }
    }

}

/*******************************************************************************
Computes an AdaBoost classifier using a set of candidate weak classifiers.
        features: Array of feature values pre-computed for the training dataset
        trainingLabel: Ground truth labels for the training examples (1 = face, 0 = background)
        numTrainingExamples: Number of training examples
        candidateWeakClassifiers: Set of candidate weak classifiers
        numCandidateWeakClassifiers: Number of candidate weak classifiers
        weakClassifiers: Set of weak classifiers selected by AdaBoost
        numWeakClassifiers: Number of selected weak classifiers
*******************************************************************************/
void MainWindow::AdaBoost(double *features, int *trainingLabel, int numTrainingExamples,
              CWeakClassifiers *candidateWeakClassifiers, int numCandidateWeakClassifiers, CWeakClassifiers *weakClassifiers, int numWeakClassifiers)
{
    FILE *out;
    out = fopen("AdaBoost.txt", "w");
    double *scores = new double [numTrainingExamples];
    double weightSum = 0.0;
    int *featureSortIdx = new int [numTrainingExamples*numCandidateWeakClassifiers];
    double *featureTranspose = new double [numTrainingExamples*numCandidateWeakClassifiers];

    // Record the classification socres for each training example
    memset(scores, 0, numTrainingExamples*sizeof(double));

    int i, j;
    // The weighting for each training example
    double *dataWeights = new double [numTrainingExamples];

    // Begin with uniform weighting
    for(i=0;i<numTrainingExamples;i++)
        dataWeights[i] = 1.0/(double) (numTrainingExamples);


    // Let's sort the feature values for each candidate weak classifier
    for(i=0;i<numCandidateWeakClassifiers;i++)
    {
        QMap<double, int> featureSort;
        QMap<double, int>::const_iterator iterator;


        for(j=0;j<numTrainingExamples;j++)
        {
            featureSort.insertMulti(features[j*numCandidateWeakClassifiers + i], j);

            // For ease later on we'll store a transposed version of the feature array
            featureTranspose[i*numTrainingExamples + j] = features[j*numCandidateWeakClassifiers + i];
        }

        j = 0;
        iterator = featureSort.constBegin();
        // Let's remember the indices of the sorted features for later.
        while (iterator != featureSort.constEnd())
        {
            featureSortIdx[i*numTrainingExamples + j] = iterator.value();
            iterator++;
            j++;
        }
    }

    // We shouldn't need the features anymore so let's delete it.
    delete [] features;


    // Find a set of weak classifiers using AdaBoost
    for(i=0;i<numWeakClassifiers;i++)
    {
        double bestError = 99999.0;
        int bestIdx = 0;

        // For each potential weak classifier
        for(j=0;j<numCandidateWeakClassifiers;j++)
        {
            CWeakClassifiers bestClassifier;

            // Find the best threshold, polarity and weight for the candidate weak classifier
            double error = FindBestClassifier(&(featureSortIdx[j*numTrainingExamples]),
                                              &(featureTranspose[j*numTrainingExamples]),
                                              trainingLabel, dataWeights, numTrainingExamples,
                                              candidateWeakClassifiers[j], &bestClassifier);

            // Is this the best classifier found so far?
            if(error < bestError)
            {
                bestError = error;
                bestIdx = j;

                // Remember the best classifier
                bestClassifier.copy(&(weakClassifiers[i]));
            }
        }

        // Given the best weak classifier found, update the weighting of the training data.
        UpdateDataWeights(&(featureTranspose[bestIdx*numTrainingExamples]), trainingLabel, weakClassifiers[i], dataWeights, numTrainingExamples);

        // Let's compute the current error for the training dataset
        weightSum += weakClassifiers[i].m_Weight;
        double error = 0.0;
        for(j=0;j<numTrainingExamples;j++)
        {
            if(featureTranspose[bestIdx*numTrainingExamples + j] > weakClassifiers[i].m_Threshold)
            {
                scores[j] += weakClassifiers[i].m_Weight*weakClassifiers[i].m_Polarity;
            }
            else
            {
                scores[j] += weakClassifiers[i].m_Weight*(1.0 - weakClassifiers[i].m_Polarity);
            }

            if((scores[j] > 0.5*weightSum && trainingLabel[j] == 0) ||
                    (scores[j] < 0.5*weightSum && trainingLabel[j] == 1))
                error++;
        }

        // Output information that you might find useful for debugging
        fprintf(out, "Count: %d\tIdx: %d\tWeight: %lf\tError: %lf\n", i, bestIdx,
                weakClassifiers[i].m_Weight, error/(double) numTrainingExamples);
        fflush(out);
    }

    delete [] dataWeights;
    delete [] scores;
    delete [] featureSortIdx;
    delete [] featureTranspose;

    fclose(out);
}

/*******************************************************************************
Find faces in an image.
        weakClassifiers: Set of weak classifiers
        numWeakClassifiers: Number of weak classifiers
        threshold: Classifier must be above Threshold to return detected face.
        minScale, maxScale: Minimum and maximum scale to search for faces.
        faceDetections: Set of face detections
        displayImage: Display image showing detected faces.
*******************************************************************************/
void MainWindow::FindFaces(QImage inImage, CWeakClassifiers *weakClassifiers, int numWeakClassifiers, double threshold, double minScale, double maxScale,
                           QMap<double, CDetection> *faceDetections, QImage *displayImage)
{
    int w = inImage.width();
    int h = inImage.height();
    double *integralImage = new double [w*h];
    double *dImage = new double [w*h];
    double scaleMulti = 1.26;
    double scale;
    int r, c;

    ConvertColorToDouble(inImage, dImage, w, h);
    // Compute the integral image
    IntegralImage(dImage, integralImage, w, h);

	double totalScore = 0.0;
	int numScores = 0;
	double minScore = 9999999999.0;
	double maxScore = 0.0;

    // Search in scale space
    for(scale=minScale;scale<maxScale;scale*=scaleMulti)
    {
        // Find size of bounding box, and the step size between neighboring bounding boxes.
        int faceSize = (int) scale;
        int stepSize = max(2, faceSize/8);



        // For every possible position
        for(r=0;r<h-faceSize;r+=stepSize)
            for(c=0;c<w-faceSize;c+=stepSize)
            {
                // Compute the score of the classifier
                double score = ClassifyBox(integralImage, c, r, faceSize, weakClassifiers, numWeakClassifiers, w);

				totalScore += score;
				numScores++;
				if(score < minScore)
				{
					minScore = score;
				}
				if(score > maxScore)
				{
					maxScore = score;
				}


                // Is the score above threshold?
                if(score > threshold)
                {
                    CDetection detection;
                    detection.m_Score = score;
                    detection.m_Scale = scale;
                    detection.m_X = (double) c;
                    detection.m_Y = (double) r;

                    // Remember the detection
                    faceDetections->insertMulti(score, detection);
                }

            }
	}

	// Draw face bounding boxes
    DrawFace(displayImage, faceDetections);

    delete [] dImage;
    delete [] integralImage;
}

/*******************************************************************************
Draw the detected faces.
        displayImage: Display image
        faceDetections: Set of face detections
*******************************************************************************/
void MainWindow::DrawFace(QImage *displayImage, QMap<double, CDetection> *faceDetections)
{
    int r, c;
    QMap<double, CDetection>::const_iterator iterator = faceDetections->constBegin();

    while(iterator != faceDetections->constEnd())
    {
        CDetection detection = iterator.value();
        int c0 = (int) detection.m_X;
        int r0 = (int) detection.m_Y;
        int size = (int) detection.m_Scale;

        for(r=r0;r<r0+size;r++)
            displayImage->setPixel(c0, r, qRgb(255, 0, 0));

        for(r=r0;r<r0+size;r++)
            displayImage->setPixel(c0 + size, r, qRgb(255, 0, 0));

        for(c=c0;c<c0+size;c++)
            displayImage->setPixel(c, r0, qRgb(255, 0, 0));

        for(c=c0;c<c0+size;c++)
            displayImage->setPixel(c, r0 + size, qRgb(255, 0, 0));

        iterator++;
    }

}


/*******************************************************************************
Display the average face and non-face image.
        displayImage: Display image, draw the average images on this image
        trainingData: Array used to store the data
        trainingLabel: Label assigned to training data (1 = face, 0 = non-face)
        numTrainingExamples: Number of training examples
        patchSize: Size of training patches in one dimension (patches have patchSize*patchSize pixels)
*******************************************************************************/
void MainWindow::DisplayAverageFace(QImage *displayImage, double *trainingData, int *trainingLabel, int numTrainingExamples, int patchSize)
{
	double * face_total = new double[patchSize * patchSize];
	double * non_face_total = new double[patchSize * patchSize];

	// Zero out total arrays
	for(int i = 0; i < patchSize * patchSize; i++)
	{
		face_total[i] = 0.0;
		non_face_total[i] = 0.0;
	}

	int num_faces = 0;
	int num_non_faces = 0;

	// Add up each of the training examples
	for(int i = 0; i < numTrainingExamples; i++)
	{
		// Counting faces
		if(trainingLabel[i])
		{
			num_faces++;
		}
		else
		{
			num_non_faces++;
		}

		// Adding face/non-face pixels
		for(int row = 0; row < patchSize; row++)
		{
			for(int col = 0; col < patchSize; col++)
			{
				if(trainingLabel[i])
				{
					face_total[row * patchSize + col] += trainingData[patchSize * patchSize * i + row * patchSize + col];
				}
				else
				{
					non_face_total[row * patchSize + col] += trainingData[patchSize * patchSize * i + row * patchSize + col];
				}
			}
		}
	}

	// Zero out displayImage
	for(int row = 0; row < displayImage->height(); row++)
	{
		for(int col = 0; col < displayImage->width(); col++)
		{
			displayImage->setPixel(col, row, qRgb(0, 0, 0));
		}
	}

	double avg_face_pixel;
	double avg_non_face_pixel;

	// Put each of the average pixels into displayImage
	for(int row = 0; row < patchSize; row++)
	{
		for(int col = 0; col < patchSize; col++)
		{
			avg_face_pixel = face_total[row * patchSize + col] / num_faces;
			avg_non_face_pixel = non_face_total[row * patchSize + col] / num_non_faces;
			displayImage->setPixel(col, row, qRgb((int)avg_face_pixel, (int)avg_face_pixel, (int)avg_face_pixel));
			displayImage->setPixel(col + patchSize, row, qRgb((int)avg_non_face_pixel, (int)avg_non_face_pixel, (int)avg_non_face_pixel));
		}
	}

	delete [] face_total;
	delete [] non_face_total;
}

/*******************************************************************************
Compute the integral image.
        image: Input double image
        integralImage: Output integral image
        w, h: Width and height of image
*******************************************************************************/
void MainWindow::IntegralImage(double *image, double *integralImage, int w, int h)
{
    double a, b, c, d;


	// Compute integral image like so:
	//
	//    ----------
	//    ----------
	//    ----------
	//    --------ab
	//    --------cd
	//
	//    d = b + c - a

	for(int row = 0; row < h; row++)
	{
		for(int col = 0; col < w; col++)
		{
			if(row == 0)
			{
				b = 0.0;
				a = 0.0;
			}
			else
			{
				b = integralImage[(row - 1) * w + col];
			}

			if(col == 0)
			{
				c = 0.0;
				a = 0.0;
			}
			else
			{
				c = integralImage[row * w + (col - 1)];
			}

			if((row != 0)&&(col != 0))
			{
				a = integralImage[(row - 1) * w + (col - 1)];
			}

			d = b + c - a;

			integralImage[row * w + col] = image[row * w + col] + d;

		}
	}
}

/*******************************************************************************
Helper function for SumBox: standard bilinear interpolation.
        image: image
        x, y: Position to interpolate
        w: Width of image (integralImage)
*******************************************************************************/
double MainWindow::BilinearInterpolation(double *image, double newX, double newY, int w)
{
	// Computing values for pixels using this setup:
	//
	//    f(x,y)              f(x+1, y)
	//      p1-------------------p2
	//      |      f(x+a, y+b)    |
	//      |            O        |
	//      |                     |
	//      |                     |
	//      p3-------------------p4
	//  f(x,y+1)              f(x+1,y+1)
	//

	int x = floor(newX);
	int y = floor(newY);
	double a = newX - x;
	double b = newY - y;

	// Get values for pixels in the diagram above
	double p1 = image[y * w + x];
	double p2 = image[y * w + (x + 1)];
	double p3 = image[(y + 1) * w + x];
	double p4 = image[(y + 1) * w + (x + 1)];

	// Use interpolation formula to calculate the value for f(x+a, y+b)
	double new_value = (1.0 - a) * (1.0 - b) * p1 + a * (1 - b) * p2 + (1 - a) * b * p3 + a * b * p4;

    return new_value;
}

/*******************************************************************************
Helper function for ComputeFeatures: compute the sum of the pixels within a box.
        integralImage: integral image
        x0, y0: Upper lefthand corner of box
        x1, y1: Lower righthand corner of box
        w: Width of image (integralImage)
*******************************************************************************/
double MainWindow::SumBox(double *integralImage, double x0, double y0, double x1, double y1, int w)
{
	
	// Calculations for box area are similar to those in IntegralImage
	//
	//      a*************b
	//      ***************
	//      ***************
	//      ***************
	//      ***************
	//      ***************
	//      c*************d
	//
	// Area = d - b - c + a
	//

	double sum = BilinearInterpolation(integralImage, x1, y1, w) - BilinearInterpolation(integralImage, x1, y0, w) -
		BilinearInterpolation(integralImage, x0, y1, w) + BilinearInterpolation(integralImage, x0, y0, w);

	return sum;

}

/*******************************************************************************
Compute all of the features for a specific bounding box.
        integralImage: integral image
        c0, r0: position of upper lefthand corner of bounding box
        size: Size of bounding box
        features: Array for storing computed feature values, access using features[i] for all i less than numWeakClassifiers.
        weakClassifiers: Weak classifiers
        numWeakClassifiers: Number of weak classifiers
        w: Width of image (integralImage)
*******************************************************************************/
void MainWindow::ComputeFeatures(double *integralImage, int c0, int r0, int size, double *features, CWeakClassifiers *weakClassifiers, int numWeakClassifiers, int w)
{
    int i, j;

    for(i = 0; i < numWeakClassifiers; i++)
    {

		features[i] = 0.0;

        for(j = 0; j < weakClassifiers[i].m_NumBoxes; j++)
        {

			// Compute the sum of the pixels within each box weakClassifiers[i].m_Box[j].
			// Multiply the sum of pixels within each box by m_BoxSign and add them together
			// to find the corresponding feature response.
			double sum = SumBox(integralImage, weakClassifiers[i].m_Box[j][0][0] * size + c0, weakClassifiers[i].m_Box[j][0][1] * size + r0,
				weakClassifiers[i].m_Box[j][1][0] * size + c0, weakClassifiers[i].m_Box[j][1][1] * size + r0, w);

            // Store the final feature value
            features[i] += weakClassifiers[i].m_BoxSign[j] * sum / ((double) (size * size));
        }
    }
}

/*******************************************************************************
AdaBoost helper function.  Find the best threshold for the candidate classifier.
        featureSortIdx: Indices of the training examples sorted based on the feature responses (lowest to highest)
                Use these indices to index into the other arrays, i.e. features, trainingLabel, dataWeights.
        features: Array of feature values for the candidate classifier
        trainingLabel: Ground truth labels for the training examples (1 = face, 0 = background)
        dataWeights: Weights used to weight each training example
        numTrainingExamples: Number of training examples
        candidateWeakClassifier: Candidate classifier
        bestClassifier: Returned best classifier (updated threshold, weight and polarity)
*******************************************************************************/
double MainWindow::FindBestClassifier(int *featureSortIdx, double *features, int *trainingLabel, double *dataWeights,
                                      int numTrainingExamples, CWeakClassifiers candidateWeakClassifier, CWeakClassifiers *bestClassifier)
{
    
	double bestError = 99999999.0;
	double bestThreshold;
	int bestPolarity;

	double totalWeight = 0.0;

	double sumWeightFaces = 0.0;
	double sumWeightBackgrounds = 0.0;
	double sumOfFacesBelowThreshold = 0.0;
	double sumOfBackgroundsBelowThreshold = 0.0;

	// Compute sums of faces and backgrounds, with weights
	for(int i = 0; i < numTrainingExamples; i++)
	{
		// If i is a background, add it to the background sum
		if(trainingLabel[featureSortIdx[i]] == 0)
		{
			sumWeightBackgrounds += dataWeights[featureSortIdx[i]];
		}
		// Otherwise, add it to the face sum
		else
		{
			sumWeightFaces += dataWeights[featureSortIdx[i]];
		}
	}

	// Try setting a threshold between every pair of features
	for(int i = 0; i < numTrainingExamples - 1; i++)
	{
		// The threshold is placed between the ith value and the i+1th value
		double thresholdValue = (features[featureSortIdx[i]] + features[featureSortIdx[i + 1]]) / 2.0;

		// If the example is a face
		if(trainingLabel[featureSortIdx[i]] == 1)
		{
			sumOfFacesBelowThreshold += dataWeights[featureSortIdx[i]];
		}
		// If the example is a background
		else
		{
			sumOfBackgroundsBelowThreshold += dataWeights[featureSortIdx[i]];
		}

		// Get minimum error for this threshold
		double error = min(sumOfFacesBelowThreshold + (sumWeightBackgrounds - sumOfBackgroundsBelowThreshold),
			sumOfBackgroundsBelowThreshold + (sumWeightFaces - sumOfFacesBelowThreshold));

		// Get the polarity corresponding to this error
		int polarity;
		polarity = int(sumOfFacesBelowThreshold + (sumWeightBackgrounds - sumOfBackgroundsBelowThreshold) <
			sumOfBackgroundsBelowThreshold + (sumWeightFaces - sumOfFacesBelowThreshold));

		// If this is a better result than we've had before,
		// store the threshold and error for this iteration
		if (error < bestError)
		{
			bestError = error;
			bestThreshold = thresholdValue;
			bestPolarity = polarity;
		}
	}

	// We now have a best error and its corresponding parameters,
	// so let's get them ready to return

    // Copy the weak classifiers params
    candidateWeakClassifier.copy(bestClassifier);

	bestClassifier->m_Polarity = bestPolarity;
	bestClassifier->m_Threshold = bestThreshold;

	// Calculate the weight from the error
	double beta = bestError / (1.0 - bestError);
	double alpha = log(1.0 / beta);
	
	bestClassifier->m_Weight = alpha;

    return bestError;
}

/*******************************************************************************
AdaBoost helper function.  Updates the weighting of the training examples.
        features: Array of feature values for the candidate classifier
        trainingLabel: Ground truth labels for the training examples (1 = face, 0 = background)
        weakClassifier: A weak classifier
        dataWeights: Weights used to weight each training example.  These are the weights updated.
        numTrainingExamples: Number of training examples
*******************************************************************************/
void MainWindow::UpdateDataWeights(double *features, int *trainingLabel, CWeakClassifiers weakClassifier, double *dataWeights, int numTrainingExamples)
{

	double alpha = weakClassifier.m_Weight;
	double beta = exp(-alpha);

	double dataWeightsTotal = 0.0;

    // Go through every example and see if this classifier predicts its "faceness" correctly
	for(int i = 0; i < numTrainingExamples; i++)
	{
		int error;	
		// If the value is over the threshold
		if(features[i] > weakClassifier.m_Threshold)
		{
			// If the polarity is 1, then it should be a face
			if(weakClassifier.m_Polarity == 1)
			{
				if(trainingLabel[i] == 1)
				{
					// Ding, correct! Adjust weight to make this node less important
					error = 0;
				}
				else
				{
					// Oops, got this wrong! Adjust weight to make this node more important
					error = 1;
				}
			}
			// Otherwise it should be a background
			else
			{
				if(trainingLabel[i] == 0)
				{
					// Ding, correct! Adjust weight to make this node less important
					error = 0;

				}
				else
				{
					// Oops, got this wrong! Adjust weight to make this node more important
					error = 1;
				}
			}
		}
		// If the value is under the threshold
		else
		{
			// If the polarity is 1, then it should be a background
			if(weakClassifier.m_Polarity == 1)
			{
				if(trainingLabel[i] == 0)
				{
					// Ding, correct! Adjust weight to make this node less important
					error = 0;

				}
				else
				{
					// Oops, got this wrong! Adjust weight to make this node more important
					error = 1;

				}
			}
			// Otherwise it should be a face
			else
			{
				if(trainingLabel[i] == 1)
				{
					// Ding, correct! Adjust weight to make this node less important
					error = 0;

				}
				else
				{
					// Oops, got this wrong! Adjust weight to make this node more important
					error = 1;
				}
			}
		}
		
		// Adjust weight
		dataWeights[i] = dataWeights[i] * pow(beta, 1 - error);
		dataWeightsTotal += dataWeights[i];
	}

	// Normalize weights
	for(int i = 0; i < numTrainingExamples; i++)
	{
		dataWeights[i] = dataWeights[i] / dataWeightsTotal;
	}
}

/*******************************************************************************
FindFaces helper function.  Return classification score for a bounding box.
        integralImage: integral image
        c0, r0: position of upper lefthand corner of bounding box
        size: Size of bounding box
        weakClassifiers: Weak classifiers
        numWeakClassifiers: Number of weak classifiers
        w: Width of image (integralImage)
*******************************************************************************/
double MainWindow::ClassifyBox(double *integralImage, int c0, int r0, int size, CWeakClassifiers *weakClassifiers, int numWeakClassifiers, int w)
{
	double * features = new double[numWeakClassifiers];
	double score = 0.0;

	ComputeFeatures(integralImage, c0, r0, size, features, weakClassifiers, numWeakClassifiers, w);

	double sum1 = 0.0;
	double sum2 = 0.0;
	int hx;

	for(int i = 0; i < numWeakClassifiers; i++)
	{
		// If the feature is greater than the threshold, it's a positive result
		if(features[i] >= weakClassifiers[i].m_Threshold)
		{
			// Positive polarity indicates it's a face
			if(weakClassifiers[i].m_Polarity == 1)
			{
				hx = 1;
			}
			// Negative polarity indicates it's a background
			else
			{
				hx = 0;
			}
		}
		// Otherwise, it's a negative result
		else
		{
			// Positive polarity indicates it's a background
			if(weakClassifiers[i].m_Polarity == 1)
			{
				hx = 0;
			}
			// Negative polarity indicates it's a face
			else
			{
				hx = 1;
			}				
		}

		// Total score using the polarities and the weights
		score += weakClassifiers[i].m_Weight * hx - 0.5 * weakClassifiers[i].m_Weight;
	}

	delete [] features;

    return score;
}

/*******************************************************************************
Non-maximal suppression of face detections (neighboring face detections must be beyond
                xyThreshold AND scaleThreshold in position and scale respectively).
        faceDetections: Set of face detections
        xyThreshold: Minimum distance in position between neighboring detections
        scaleThreshold: Minimum distance in scale between neighboring detections
        displayImage: Display image
*******************************************************************************/
void MainWindow::NMS(QMap<double, CDetection> *faceDetections, double xyThreshold, double scaleThreshold, QImage *displayImage)
{
    QMap<double, CDetection>::const_iterator iterator = faceDetections->constBegin();
	QMap<double, CDetection>::const_iterator iterator2;

	// First copy faceDetections so we're not editing the map we're iterating over
	QMap<double, CDetection> finalFaceDetections = *faceDetections;
    
	double previous;
	bool first, previousRemoved;

	// Iterate through all the faces detections (lowest face detection score first)
    while(iterator != faceDetections->constEnd())
    {
		// Compare one value to each value above it
		iterator2 = iterator + 1;
		while(iterator2 != faceDetections->constEnd())
		{
			// If the distance between the two boxes is within xyThreshold, and the scale difference is within scaleThreshold
			if((sqrt((iterator.value().m_X - iterator2.value().m_X) * (iterator.value().m_X - iterator2.value().m_X)
					+ (iterator.value().m_Y - iterator2.value().m_Y) * (iterator.value().m_Y - iterator2.value().m_Y)) < xyThreshold)
				&& (sqrt((iterator.value().m_Scale - iterator2.value().m_Scale) * (iterator.value().m_Scale - iterator2.value().m_Scale)) < scaleThreshold))
			{
				// Remove the one of the two with the lesser score
				if(iterator.value().m_Score < iterator2.value().m_Score)
				{
					finalFaceDetections.remove(iterator.key());
				}
				else
				{
					finalFaceDetections.remove(iterator2.key());
					continue;
				}
			}
			iterator2++;
		}

        // Add a face detection to finalFaceDetections using:
        // finalFaceDetections.insertMulti(iterator.key(), iterator.value());

        iterator++;
    }

	// Add all remaining members of faceDetections to finalFaceDetections
	DrawFace(displayImage, &finalFaceDetections);
}